from __future__ import division
from __future__ import print_function

import keras

from keras.applications import resnet50, densenet, inception_v3, inception_resnet_v2, mobilenetv2, nasnet, xception
from keras.preprocessing.image import ImageDataGenerator
from keras.utils import multi_gpu_model
from collections import OrderedDict
from math import ceil
from skimage.io import imread
from skimage.transform import resize
import numpy as np
import cv2
import os
import json
import csv


from ....lib.generators import PredictGenerator, TrainGenerator
from ....lib.general_lib import get_n_largest_index_and_probs, get_data_path_type, reverse_dict
from ....lib.image_lib import directory_images_to_arrays, reshape_data
from ....lib.saving_lib import save_models_with_class_indices, load_models_with_class_indices
from ....lib.custom_callback import ImageClassificationCallback
from ....lib.model_lib import image_classification_model
from ....lib.config_lib import ImageClassificationConfig, DeepblockLogConfig

from ...application import Application


class Image_Classification(Application):
    def __init__(self, model_path=None, model_type = 'densenet121'):
        super().__init__()
        
        if not isinstance(model_path, type(None)):
            self.load(model_path)
        else:
            self.model_type = model_type

    def train(self,
            data_path=None,
            optimizer='nadam',
            metrics=['accuracy'],
            epochs=1,
            batch_size=32,
            verbose = 0,
            shuffle = False):
        get_image_from = get_data_path_type(data_path)

        train_generator = self.make_train_generator(get_image_from = get_image_from,
                                                data_path = data_path,
                                                batch_size = batch_size)
        class_indices = self.change_dict_value_int_from_string(train_generator.class_indices)
        num_dataset = train_generator.n
        steps_per_epoch = ceil(num_dataset / batch_size)

        if not hasattr(self, 'model'):
            self.model = self.create_and_compile_model(class_indices = class_indices,
                                                model_type = self.model_type,
                                                optimizer = optimizer,
                                                metrics = metrics)
        
        callbacks = []
        if self.deepblock_log:
            callbacks = [ImageClassificationCallback(
                total_step=steps_per_epoch, total_epoch=epochs)]
        self.model.fit_generator(
            train_generator,
            steps_per_epoch=steps_per_epoch,
            epochs=epochs,
            verbose=verbose,
            callbacks=callbacks,
            shuffle=shuffle)

    def predict(self, data_path, batch_size=16, verbose=0):
        data_path_type = get_data_path_type(data_path)

        pair_dict = OrderedDict()
        if data_path_type == 'directory':

            datagen = One_Directory_Image_Generator()
            predict_data_generator = datagen.flow_one_directory(
                    directory_path=data_path,
                    shuffle=False,
                    batch_size=batch_size,
                    image_shape=self.model.input_shape[1:3]
                )
            
            num_dataset = len(predict_data_generator.image_filenames)
            probs = self.model.predict_generator(
                predict_data_generator, steps=ceil(num_dataset / batch_size))
            predicted_order_filenames = predict_data_generator.predict_order_filenames
            predicted_class_indices = probs.argmax(axis=-1)

            reversed_dict = reverse_dict(self.model.class_indices)
            for i in range(len(predicted_class_indices)):
                pair_dict[predicted_order_filenames[i]] = reversed_dict[str(probs.argmax(axis=-1)[i])]

            if self.deepblock_log == True:
                log_dict = OrderedDict()
                log_dict[DeepblockLogConfig.LOG_TYPE] = DeepblockLogConfig.PREDICT_END
                log_dict[DeepblockLogConfig.RESULT] = pair_dict
                return json.dumps(log_dict, ensure_ascii=False)
            else:
                return json.dumps(pair_dict, ensure_ascii=False)
        elif data_path_type == 'image':
            _, only_file_name = os.path.split(data_path)
            img_to_predict = cv2.imread(data_path)
            img_array = np.expand_dims(img_to_predict, axis=0)
            reshaped_array = reshape_data(data_array = img_array, 
                                        input_shape = self.model.input_shape)
            probs = self.model.predict(
                reshaped_array,
                batch_size=batch_size,
                verbose=verbose)

            reversed_dict = reverse_dict(self.model.class_indices)
            pair_dict[only_file_name] = reversed_dict[str(probs.argmax(axis=-1)[0])]

            if self.deepblock_log == True:
                log_dict = OrderedDict()
                log_dict[DeepblockLogConfig.LOG_TYPE] = DeepblockLogConfig.PREDICT_END
                log_dict[DeepblockLogConfig.RESULT] = pair_dict
                return json.dumps(log_dict, ensure_ascii=False)
            else:
                return json.dumps(pair_dict, ensure_ascii=False)

    def load(self, model_path=None):
        try:
            self.model = load_models_with_class_indices(model_path)
            if self.gpu_num > 1:
                self.model = multi_gpu_model(self.model, gpus = self.gpu_num)
        except Exception as e:
            raise e


    def save(self, model_path):
        try:
            if isinstance(self.model, type(None)):
                raise TypeError('You should create a model before saving it')

            if self.deepblock_log == True:
                logJson = OrderedDict()
                logJson[DeepblockLogConfig.LOG_TYPE] = DeepblockLogConfig.SAVE_START
                logJson[DeepblockLogConfig.SUCCESS] = "true"
                print(json.dumps(logJson, ensure_ascii=False))

            save_models_with_class_indices(self.model, model_path)

            if self.deepblock_log == True:
                logJson = OrderedDict()
                logJson[DeepblockLogConfig.LOG_TYPE] = DeepblockLogConfig.SAVE_END
                logJson[DeepblockLogConfig.SUCCESS] = "true"
                print(json.dumps(logJson, ensure_ascii=False))
        except Exception as e:
            raise e


    def create_and_compile_model(self, class_indices, model_type, optimizer, metrics):
        model = image_classification_model(num_classes = len(class_indices), 
                                        gpu_num = self.gpu_num,
                                        model_type = model_type)
        model.compile(optimizer=optimizer, 
                    loss = keras.losses.categorical_crossentropy, 
                    metrics=metrics)
        model.class_indices = class_indices
        return model


    def make_train_generator(self, get_image_from, data_path, batch_size):
        if get_image_from == 'directory':
            train_datagen = ImageDataGenerator(rescale = 1. / 255.)
            train_generator = train_datagen.flow_from_directory(
                                                    data_path, 
                                                    target_size=ImageClassificationConfig.INPUT_SHAPE[0:2],
                                                    class_mode='categorical', 
                                                    batch_size=batch_size)
        elif get_image_from == 'csv':
            train_datagen = CSV_Generator()
            train_generator = train_datagen.flow_from_csv(csv_file_path=data_path,
                                                          openimage_directory_path="",
                                                          target_size=ImageClassificationConfig.INPUT_SHAPE[0:2],
                                                          batch_size=batch_size)

        return train_generator


    def evaluate(self, data_path=None, batch_size=1):
        try:
            if isinstance(data_path, type(None)):
                raise TypeError('You sholud give a proper data path')

            evaluate_datagen = ImageDataGenerator(rescale=1. / 255)
            validation_generator = evaluate_datagen.flow_from_directory(
                data_path,
                target_size=self.model.input_shape[1:3],
                batch_size=batch_size
            )

            scores = self.model.evaluate_generator(
                validation_generator, 
                ceil(len(validation_generator) / batch_size))
            print("accuracy : ", scores[1] * 100)
        except Exception as e:
            raise e

    
    def change_dict_value_int_from_string(self, class_indices):
        string_class_indices = {}
        for key, value in class_indices.items():
            string_class_indices[key] = str(value)
        
        return string_class_indices

    
class CSV_Generator(keras.utils.Sequence):
    def __init__(self, image_filenames=None, image_shape=None, shuffle=False):
        self.image_filenames = image_filenames
        self.image_classes = []
        self.class_indices = []
        self.image_shape = image_shape
        self.shuffle = shuffle

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_x = self.image_filenames[idx *
                                       self.batch_size:(idx + 1) * self.batch_size]
        class_x = self.image_classes[idx *
                                     self.batch_size:(idx + 1) * self.batch_size]

        cnt = 0
        x = []
        index_list = []
        for filename in batch_x:
            try:
                img = cv2.imread(filename)
                resized_img = cv2.resize(img, self.image_shape)
                x.append(resized_img)
                index_list.append(cnt)
                cnt += 1
            except BaseException:
                pass

        batch_y = np.zeros((cnt, len(self.class_indices)), dtype='float')

        i = 0
        for index in index_list:
            find_index = self.class_indices.index(class_x[index])
            batch_y[i, find_index] = 1
            i += 1

        x_array = np.array(x)
        return x_array, batch_y

    def flow_from_csv(self, csv_file_path, openimage_directory_path,
                      shuffle=False, target_size=None, batch_size=1):
        f = open(csv_file_path, 'r', encoding='utf-8')
        csv_reader = csv.reader(f)
        self.image_classes = list()
        self.image_filenames = list()

        for line in csv_reader:
            atoms = line
            file_full_path = os.path.join(
                openimage_directory_path, atoms[0] + '.jpg')
            if os.path.isfile(file_full_path):
                self.image_classes.append(atoms[2])
                self.image_filenames.append(file_full_path)
        f.close()
        self.class_indices = list(set(self.image_classes))
        self.shuffle = shuffle
        self.image_shape = target_size
        self.batch_size = batch_size
        self.n = len(self.image_filenames)
        return self

class One_Directory_Image_Generator(keras.utils.Sequence):
    def __init__(self, image_filenames=None, image_classes=None,
                 batch_size=None, image_shape=None, shuffle=False):
        self.image_filenames = image_filenames
        self.image_classes = image_classes
        self.batch_size = batch_size
        self.image_shape = image_shape
        self.shuffle = shuffle
        self.predict_order_filenames = []

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_x = self.image_filenames[idx *
                                       self.batch_size:(idx + 1) * self.batch_size]

        x = []
        for filename in batch_x:
            _, only_file_name = os.path.split(filename)
            self.predict_order_filenames.append(only_file_name)
            img = cv2.imread(filename)
            test_data = np.expand_dims(img, axis=0)
            resized_img = cv2.resize(test_data[0], self.image_shape)
            x.append(resized_img)

        data_array = np.asarray(x)
        data_array = data_array.astype('float32')
        data_array /= 255
        return data_array

    def flow_one_directory(self, directory_path,
                           shuffle=False, batch_size=1, image_shape=None):
        filelist = os.listdir(directory_path)
        for i in range(len(filelist)):
            if i == 0:
                self.image_filenames = [directory_path + '/' + filelist[i]]
            else:
                self.image_filenames.append(directory_path + '/' + filelist[i])
        self.batch_size = batch_size
        self.image_shape = image_shape
        self.shuffle = shuffle

        return self



"""
Mask R-CNN
Display and Visualization Functions.

Copyright (c) 2017 Matterport, Inc.
Licensed under the MIT License (see LICENSE for details)
Written by Waleed Abdulla
"""

from . import utils
import os
import sys
import random
import itertools
import colorsys

import numpy as np
from skimage.measure import find_contours
import matplotlib.pyplot as plt
from matplotlib import patches, lines
from matplotlib.patches import Polygon

# Root directory of the project
ROOT_DIR = os.path.abspath("../")

# Import Mask RCNN
sys.path.append(ROOT_DIR)  # To find local version of the library


############################################################
#  Visualization
############################################################

def display_images(images, titles=None, cols=4, cmap=None, norm=None,
                   interpolation=None):
    """Display the given set of images, optionally with titles.
    images: list or array of image tensors in HWC format.
    titles: optional. A list of titles to display with each image.
    cols: number of images per row
    cmap: Optional. Color map to use. For example, "Blues".
    norm: Optional. A Normalize instance to map values to colors.
    interpolation: Optional. Image interpolation to use for display.
    """
    titles = titles if titles is not None else [""] * len(images)
    rows = len(images) // cols + 1
    plt.figure(figsize=(14, 14 * rows // cols))
    i = 1
    for image, title in zip(images, titles):
        plt.subplot(rows, cols, i)
        plt.title(title, fontsize=9)
        plt.axis('off')
        plt.imshow(image.astype(np.uint8), cmap=cmap,
                   norm=norm, interpolation=interpolation)
        i += 1
    plt.show()


def random_colors(N, bright=True):
    """
    Generate random colors.
    To get visually distinct colors, generate them in HSV space then
    convert to RGB.
    """
    brightness = 1.0 if bright else 0.7
    hsv = [(i / N, 1, brightness) for i in range(N)]
    colors = list(map(lambda c: colorsys.hsv_to_rgb(*c), hsv))
    random.shuffle(colors)
    return colors


def apply_mask(image, mask, color, alpha=0.5):
    """Apply the given mask to the image.
    """
    for c in range(3):
        image[:, :, c] = np.where(mask == 1,
                                  image[:, :, c] *
                                  (1 - alpha) + alpha * color[c] * 255,
                                  image[:, :, c])
    return image


def display_instances(image, boxes, masks, class_ids, class_names,
                      scores=None, title="",
                      figsize=(16, 16), ax=None,
                      show_mask=True, show_bbox=True,
                      colors=None, captions=None, **kwargs):
    """
    boxes: [num_instance, (y1, x1, y2, x2, class_id)] in image coordinates.
    masks: [height, width, num_instances]
    class_ids: [num_instances]
    class_names: list of class names of the dataset
    scores: (optional) confidence scores for each box
    title: (optional) Figure title
    show_mask, show_bbox: To show masks and bounding boxes or not
    figsize: (optional) the size of the image
    colors: (optional) An array or colors to use with each object
    captions: (optional) A list of strings to use as captions for each object
    """
    # Number of instances
    N = boxes.shape[0]
    if not N:
        print("\n*** No instances to display *** \n")
    else:
        assert boxes.shape[0] == masks.shape[-1] == class_ids.shape[0]

    # If no axis is passed, create one and automatically call show()
    auto_show = False
    if not ax:
        _, ax = plt.subplots(1, figsize=figsize)
        auto_show = True

    # Generate random colors
    colors = colors or random_colors(N)

    # Show area outside image boundaries.
    height, width = image.shape[:2]
    ax.set_ylim(height + 10, -10)
    ax.set_xlim(-10, width + 10)
    ax.axis('off')
    ax.set_title(title)

    masked_image = image.astype(np.uint32).copy()
    contours_list = []
    for i in range(N):
        color = colors[i]

        # Bounding box
        if not np.any(boxes[i]):
            # Skip this instance. Has no bbox. Likely lost in image cropping.
            continue
        y1, x1, y2, x2 = boxes[i]
        if show_bbox:
            p = patches.Rectangle((x1, y1), x2 - x1, y2 - y1, linewidth=2,
                                  alpha=0.7, linestyle="dashed",
                                  edgecolor=color, facecolor='none')
            ax.add_patch(p)

        # Label
        if not captions:
            class_id = class_ids[i]
            score = scores[i] if scores is not None else None
            label = class_names[class_id]
            x = random.randint(x1, (x1 + x2) // 2)
            caption = "{} {:.3f}".format(label, score) if score else label
        else:
            caption = captions[i]
        ax.text(x1, y1 + 8, caption,
                color='w', size=11, backgroundcolor="none")

        # Mask
        mask = masks[:, :, i]
        if show_mask:
            masked_image = apply_mask(masked_image, mask, color)

        # Mask Polygon
        # Pad to ensure proper polygons for masks that touch image edges.
        padded_mask = np.zeros(
            (mask.shape[0] + 2, mask.shape[1] + 2), dtype=np.uint8)
        padded_mask[1:-1, 1:-1] = mask
        contours = find_contours(padded_mask, 0.5)
        for verts in contours:
            # Subtract the padding and flip (y, x) to (x, y)
            verts = np.fliplr(verts) - 1
            contours_list.append(verts)
            p = Polygon(verts, facecolor="none", edgecolor=color)
            ax.add_patch(p)
    ax.imshow(masked_image.astype(np.uint8))
    # if auto_show or show_masked_image:
    #     plt.show()
    return plt, contours_list


def display_differences(image,
                        gt_box, gt_class_id, gt_mask,
                        pred_box, pred_class_id, pred_score, pred_mask,
                        class_names, title="", ax=None,
                        show_mask=True, show_box=True,
                        iou_threshold=0.5, score_threshold=0.5):
    """Display ground truth and prediction instances on the same image."""
    # Match predictions to ground truth
    gt_match, pred_match, overlaps = utils.compute_matches(
        gt_box, gt_class_id, gt_mask,
        pred_box, pred_class_id, pred_score, pred_mask,
        iou_threshold=iou_threshold, score_threshold=score_threshold)
    # Ground truth = green. Predictions = red
    colors = [(0, 1, 0, .8)] * len(gt_match)\
        + [(1, 0, 0, 1)] * len(pred_match)
    # Concatenate GT and predictions
    class_ids = np.concatenate([gt_class_id, pred_class_id])
    scores = np.concatenate([np.zeros([len(gt_match)]), pred_score])
    boxes = np.concatenate([gt_box, pred_box])
    masks = np.concatenate([gt_mask, pred_mask], axis=-1)
    # Captions per instance show score/IoU
    captions = ["" for m in gt_match] + ["{:.2f} / {:.2f}".format(
        pred_score[i],
        (overlaps[i, int(pred_match[i])]
            if pred_match[i] > -1 else overlaps[i].max()))
        for i in range(len(pred_match))]
    # Set title if not provided
    title = title or "Ground Truth and Detections\n GT=green, pred=red, captions: score/IoU"
    # Display
    display_instances(
        image,
        boxes, masks, class_ids,
        class_names, scores, ax=ax,
        show_bbox=show_box, show_mask=show_mask,
        colors=colors, captions=captions,
        title=title)


def draw_rois(image, rois, refined_rois, mask,
              class_ids, class_names, limit=10):
    """
    anchors: [n, (y1, x1, y2, x2)] list of anchors in image coordinates.
    proposals: [n, 4] the same anchors but refined to fit objects better.
    """
    masked_image = image.copy()

    # Pick random anchors in case there are too many.
    ids = np.arange(rois.shape[0], dtype=np.int32)
    ids = np.random.choice(
        ids, limit, replace=False) if ids.shape[0] > limit else ids

    fig, ax = plt.subplots(1, figsize=(12, 12))
    if rois.shape[0] > limit:
        plt.title("Showing {} random ROIs out of {}".format(
            len(ids), rois.shape[0]))
    else:
        plt.title("{} ROIs".format(len(ids)))

    # Show area outside image boundaries.
    ax.set_ylim(image.shape[0] + 20, -20)
    ax.set_xlim(-50, image.shape[1] + 20)
    ax.axis('off')

    for i, id in enumerate(ids):
        color = np.random.rand(3)
        class_id = class_ids[id]
        # ROI
        y1, x1, y2, x2 = rois[id]
        p = patches.Rectangle((x1, y1), x2 - x1, y2 - y1, linewidth=2,
                              edgecolor=color if class_id else "gray",
                              facecolor='none', linestyle="dashed")
        ax.add_patch(p)
        # Refined ROI
        if class_id:
            ry1, rx1, ry2, rx2 = refined_rois[id]
            p = patches.Rectangle(
                (rx1,
                 ry1),
                rx2 - rx1,
                ry2 - ry1,
                linewidth=2,
                edgecolor=color,
                facecolor='none')
            ax.add_patch(p)
            # Connect the top-left corners of the anchor and proposal for easy
            # visualization
            ax.add_line(lines.Line2D([x1, rx1], [y1, ry1], color=color))

            # Label
            label = class_names[class_id]
            ax.text(rx1, ry1 + 8, "{}".format(label),
                    color='w', size=11, backgroundcolor="none")

            # Mask
            m = utils.unmold_mask(mask[id], rois[id]
                                  [:4].astype(np.int32), image.shape)
            masked_image = apply_mask(masked_image, m, color)

    ax.imshow(masked_image)

    # Print stats
    print("Positive ROIs: ", class_ids[class_ids > 0].shape[0])
    print("Negative ROIs: ", class_ids[class_ids == 0].shape[0])
    print("Positive Ratio: {:.2f}".format(
        class_ids[class_ids > 0].shape[0] / class_ids.shape[0]))


# TODO: Replace with matplotlib equivalent?
def draw_box(image, box, color):
    """Draw 3-pixel width bounding boxes on the given image array.
    color: list of 3 int values for RGB.
    """
    y1, x1, y2, x2 = box
    image[y1:y1 + 2, x1:x2] = color
    image[y2:y2 + 2, x1:x2] = color
    image[y1:y2, x1:x1 + 2] = color
    image[y1:y2, x2:x2 + 2] = color
    return image


def display_top_masks(image, mask, class_ids, class_names, limit=4):
    """Display the given image and the top few class masks."""
    to_display = []
    titles = []
    to_display.append(image)
    titles.append("H x W={}x{}".format(image.shape[0], image.shape[1]))
    # Pick top prominent classes in this image
    unique_class_ids = np.unique(class_ids)
    mask_area = [np.sum(mask[:, :, np.where(class_ids == i)[0]])
                 for i in unique_class_ids]
    top_ids = [v[0] for v in sorted(zip(unique_class_ids, mask_area),
                                    key=lambda r: r[1], reverse=True) if v[1] > 0]
    # Generate images and titles
    for i in range(limit):
        class_id = top_ids[i] if i < len(top_ids) else -1
        # Pull masks of instances belonging to the same class.
        m = mask[:, :, np.where(class_ids == class_id)[0]]
        m = np.sum(m * np.arange(1, m.shape[-1] + 1), -1)
        to_display.append(m)
        titles.append(class_names[class_id] if class_id != -1 else "-")
    display_images(to_display, titles=titles, cols=limit + 1, cmap="Blues_r")


def plot_precision_recall(AP, precisions, recalls):
    """Draw the precision-recall curve.

    AP: Average precision at IoU >= 0.5
    precisions: list of precision values
    recalls: list of recall values
    """
    # Plot the Precision-Recall curve
    _, ax = plt.subplots(1)
    ax.set_title("Precision-Recall Curve. AP@50 = {:.3f}".format(AP))
    ax.set_ylim(0, 1.1)
    ax.set_xlim(0, 1.1)
    _ = ax.plot(recalls, precisions)


def plot_overlaps(gt_class_ids, pred_class_ids, pred_scores,
                  overlaps, class_names, threshold=0.5):
    """Draw a grid showing how ground truth objects are classified.
    gt_class_ids: [N] int. Ground truth class IDs
    pred_class_id: [N] int. Predicted class IDs
    pred_scores: [N] float. The probability scores of predicted classes
    overlaps: [pred_boxes, gt_boxes] IoU overlaps of predictions and GT boxes.
    class_names: list of all class names in the dataset
    threshold: Float. The prediction probability required to predict a class
    """
    gt_class_ids = gt_class_ids[gt_class_ids != 0]
    pred_class_ids = pred_class_ids[pred_class_ids != 0]

    plt.figure(figsize=(12, 10))
    plt.imshow(overlaps, interpolation='nearest', cmap=plt.cm.Blues)
    plt.yticks(np.arange(len(pred_class_ids)),
               ["{} ({:.2f})".format(class_names[int(id)], pred_scores[i])
                for i, id in enumerate(pred_class_ids)])
    plt.xticks(np.arange(len(gt_class_ids)),
               [class_names[int(id)] for id in gt_class_ids], rotation=90)

    thresh = overlaps.max() / 2.
    for i, j in itertools.product(range(overlaps.shape[0]),
                                  range(overlaps.shape[1])):
        text = ""
        if overlaps[i, j] > threshold:
            text = "match" if gt_class_ids[j] == pred_class_ids[i] else "wrong"
        color = ("white" if overlaps[i, j] > thresh
                 else "black" if overlaps[i, j] > 0
                 else "grey")
        plt.text(j, i, "{:.3f}\n{}".format(overlaps[i, j], text),
                 horizontalalignment="center", verticalalignment="center",
                 fontsize=9, color=color)

    plt.tight_layout()
    plt.xlabel("Ground Truth")
    plt.ylabel("Predictions")


def draw_boxes(image, boxes=None, refined_boxes=None,
               masks=None, captions=None, visibilities=None,
               title="", ax=None):
    """Draw bounding boxes and segmentation masks with different
    customizations.

    boxes: [N, (y1, x1, y2, x2, class_id)] in image coordinates.
    refined_boxes: Like boxes, but draw with solid lines to show
        that they're the result of refining 'boxes'.
    masks: [N, height, width]
    captions: List of N titles to display on each box
    visibilities: (optional) List of values of 0, 1, or 2. Determine how
        prominent each bounding box should be.
    title: An optional title to show over the image
    ax: (optional) Matplotlib axis to draw on.
    """
    # Number of boxes
    assert boxes is not None or refined_boxes is not None
    N = boxes.shape[0] if boxes is not None else refined_boxes.shape[0]

    # Matplotlib Axis
    if not ax:
        _, ax = plt.subplots(1, figsize=(12, 12))

    # Generate random colors
    colors = random_colors(N)

    # Show area outside image boundaries.
    margin = image.shape[0] // 10
    ax.set_ylim(image.shape[0] + margin, -margin)
    ax.set_xlim(-margin, image.shape[1] + margin)
    ax.axis('off')

    ax.set_title(title)

    masked_image = image.astype(np.uint32).copy()
    for i in range(N):
        # Box visibility
        visibility = visibilities[i] if visibilities is not None else 1
        if visibility == 0:
            color = "gray"
            style = "dotted"
            alpha = 0.5
        elif visibility == 1:
            color = colors[i]
            style = "dotted"
            alpha = 1
        elif visibility == 2:
            color = colors[i]
            style = "solid"
            alpha = 1

        # Boxes
        if boxes is not None:
            if not np.any(boxes[i]):
                # Skip this instance. Has no bbox. Likely lost in cropping.
                continue
            y1, x1, y2, x2 = boxes[i]
            p = patches.Rectangle((x1, y1), x2 - x1, y2 - y1, linewidth=2,
                                  alpha=alpha, linestyle=style,
                                  edgecolor=color, facecolor='none')
            ax.add_patch(p)

        # Refined boxes
        if refined_boxes is not None and visibility > 0:
            ry1, rx1, ry2, rx2 = refined_boxes[i].astype(np.int32)
            p = patches.Rectangle(
                (rx1,
                 ry1),
                rx2 - rx1,
                ry2 - ry1,
                linewidth=2,
                edgecolor=color,
                facecolor='none')
            ax.add_patch(p)
            # Connect the top-left corners of the anchor and proposal
            if boxes is not None:
                ax.add_line(lines.Line2D([x1, rx1], [y1, ry1], color=color))

        # Captions
        if captions is not None:
            caption = captions[i]
            # If there are refined boxes, display captions on them
            if refined_boxes is not None:
                y1, x1, y2, x2 = ry1, rx1, ry2, rx2
            x = random.randint(x1, (x1 + x2) // 2)
            ax.text(x1, y1, caption, size=11, verticalalignment='top',
                    color='w', backgroundcolor="none",
                    bbox={'facecolor': color, 'alpha': 0.5,
                          'pad': 2, 'edgecolor': 'none'})

        # Masks
        if masks is not None:
            mask = masks[:, :, i]
            masked_image = apply_mask(masked_image, mask, color)
            # Mask Polygon
            # Pad to ensure proper polygons for masks that touch image edges.
            padded_mask = np.zeros(
                (mask.shape[0] + 2, mask.shape[1] + 2), dtype=np.uint8)
            padded_mask[1:-1, 1:-1] = mask
            contours = find_contours(padded_mask, 0.5)
            for verts in contours:
                # Subtract the padding and flip (y, x) to (x, y)
                verts = np.fliplr(verts) - 1
                p = Polygon(verts, facecolor="none", edgecolor=color)
                ax.add_patch(p)
    ax.imshow(masked_image.astype(np.uint8))


def display_table(table):
    """Display values in a table format.
    table: an iterable of rows, and each row is an iterable of values.
    """
    html = ""
    for row in table:
        row_html = ""
        for col in row:
            row_html += "<td>{:40}</td>".format(str(col))
        html += "<tr>" + row_html + "</tr>"
    html = "<table>" + html + "</table>"
    # IPython.display.display(IPython.display.HTML(html))


def display_weight_stats(model):
    """Scans all the weights in the model and returns a list of tuples
    that contain stats about each weight.
    """
    layers = model.get_trainable_layers()
    table = [["WEIGHT NAME", "SHAPE", "MIN", "MAX", "STD"]]
    for l in layers:
        weight_values = l.get_weights()  # list of Numpy arrays
        weight_tensors = l.weights  # list of TF tensors
        for i, w in enumerate(weight_values):
            weight_name = weight_tensors[i].name
            # Detect problematic layers. Exclude biases of conv layers.
            alert = ""
            if w.min() == w.max() and not (l.__class__.__name__ == "Conv2D" and i == 1):
                alert += "<span style='color:red'>*** dead?</span>"
            if np.abs(w.min()) > 1000 or np.abs(w.max()) > 1000:
                alert += "<span style='color:red'>*** Overflow?</span>"
            # Add row
            table.append([
                weight_name + alert,
                str(w.shape),
                "{:+9.4f}".format(w.min()),
                "{:+10.4f}".format(w.max()),
                "{:+9.4f}".format(w.std()),
            ])
    display_table(table)



from keras.layers import Input, Activation, Add, UpSampling2D, Dropout, MaxPooling2D, PReLU
from keras.layers.advanced_activations import LeakyReLU
from keras.layers.convolutional import Conv2D
from keras.layers.core import Dense, Flatten, Lambda
from keras.layers.normalization import BatchNormalization
from keras.models import Model, Sequential
from keras import optimizers, losses
from keras.utils import multi_gpu_model
from keras.applications import resnet50, densenet, inception_v3, inception_resnet_v2, mobilenetv2, nasnet, xception
from .layers_lib import ReflectionPadding2D, res_block
from .config_lib import DeblurConfig, MultiLabelClassificationConfig, ImageClassificationConfig


####################IMAGE_CLASSIFICATION_MODEL####################
def image_classification_model(num_classes, gpu_num, model_type):
    if model_type == 'densenet121':
        model = densenet.DenseNet121(weights=None, classes=num_classes, input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'inception_v3':
        model = inception_v3.InceptionV3(weights=None, classes=num_classes, input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'inception_resnet_v2':
        model = inception_v3.InceptionV3(weights=None, classes=num_classes, input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'mobilenet_v2':
        model = mobilenetv2.MobileNetV2(weights=None, classes=num_classes, input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'nasnet_large':
        model = nasnet.NASNetLarge(weights=None, classes=num_classes, input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'nasnet_mobile':
        model = nasnet.NASNetMobile(weights=None, classes=num_classes, input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'xception':
        model = xception.Xception(weights=None, classes=num_classes, input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'resnet50':
        model = resnet50.ResNet50(weights=None, classes=num_classes, input_shape=ImageClassificationConfig.INPUT_SHAPE)

    if gpu_num > 1:
        model = multi_gpu_model(model, gpus = gpu_num)

    return model


####################MULTI_LABEL_IMAGE_CLASSIFICATION_MODEL####################
def multi_label_classification_model(num_classes, gpu_num = 1, learning_rate = 0.0001):
    model = Sequential()
    model.add(Conv2D(32, (3, 3), padding='same',
                     input_shape=MultiLabelClassificationConfig.INPUT_SHAPE))
    model.add(Activation('relu'))
    model.add(Conv2D(32, (3, 3)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Dropout(0.25))
    model.add(Conv2D(64, (3, 3), padding='same'))
    model.add(Activation('relu'))
    model.add(Conv2D(64, (3, 3)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Dropout(0.25))
    model.add(Flatten())
    model.add(Dense(512))
    model.add(Activation('relu'))
    model.add(Dropout(0.5))
    model.add(Dense(num_classes, activation='sigmoid'))

    if gpu_num > 1:
        model = multi_gpu_model(model, gpus = gpu_num)

    return model

def multi_label_classification_model_2(num_classes, gpu_num=1, learning_rate = 0.001):
    model = Sequential()
    model.add(Conv2D(32, (3, 3), padding="same",
                     input_shape=MultiLabelClassificationConfig.INPUT_SHAPE))
    model.add(Activation("relu"))
    model.add(BatchNormalization(axis=-1))
    model.add(MaxPooling2D(pool_size=(3, 3)))
    model.add(Dropout(0.25))

    model.add(Conv2D(64, (3, 3), padding="same"))
    model.add(Activation("relu"))
    model.add(BatchNormalization(axis=-1))
    model.add(Conv2D(64, (3, 3), padding="same"))
    model.add(Activation("relu"))
    model.add(BatchNormalization(axis=-1))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Dropout(0.25))

    model.add(Flatten())
    model.add(Dense(1024))
    model.add(Activation("relu"))
    model.add(BatchNormalization())
    model.add(Dropout(0.5))

    model.add(Dense(num_classes))
    model.add(Activation("sigmoid"))

    if gpu_num > 1:
        model = multi_gpu_model(model, gpus = gpu_num)

    return model


####################DEBLUR_GAN_MODELS####################
def generator_model():
    ngf = DeblurConfig.NGF
    n_blocks_gen = DeblurConfig.N_BLOCKS_GEN
    shape = DeblurConfig.CELL_SHAPE
    output_nc = DeblurConfig.OUTPUT_NC

    """Build generator architecture."""
    # Current version : ResNet block
    inputs = Input(shape=shape)

    x = ReflectionPadding2D((3, 3))(inputs)
    x = Conv2D(filters=ngf, kernel_size=(7, 7), padding='valid')(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)

    n_downsampling = 2
    for i in range(n_downsampling):
        mult = 2**i
        x = Conv2D(filters=ngf * mult * 2, kernel_size=(3, 3),
                   strides=2, padding='same')(x)
        x = BatchNormalization()(x)
        x = Activation('relu')(x)

    mult = 2**n_downsampling
    for i in range(n_blocks_gen):
        x = res_block(x, ngf * mult, use_dropout=True)

    for i in range(n_downsampling):
        mult = 2**(n_downsampling - i)
        # x = Conv2DTranspose(filters=int(ngf * mult / 2), kernel_size=(3, 3), strides=2, padding='same')(x)
        x = UpSampling2D()(x)
        x = Conv2D(filters=int(ngf * mult / 2),
                   kernel_size=(3, 3), padding='same')(x)
        x = BatchNormalization()(x)
        x = Activation('relu')(x)

    x = ReflectionPadding2D((3, 3))(x)
    x = Conv2D(filters=output_nc, kernel_size=(7, 7), padding='valid')(x)
    x = Activation('tanh')(x)

    outputs = Add()([x, inputs])
    # outputs = Lambda(lambda z: K.clip(z, -1, 1))(x)
    outputs = Lambda(lambda z: z / 2)(outputs)

    model = Model(inputs=inputs, outputs=outputs, name='Generator')
    return model


def discriminator_model():
    ndf = DeblurConfig.NDF
    shape = DeblurConfig.CELL_SHAPE
    """Build discriminator architecture."""
    n_layers, use_sigmoid = 3, False
    inputs = Input(shape=shape)

    x = Conv2D(filters=ndf, kernel_size=(4, 4),
               strides=2, padding='same')(inputs)
    x = LeakyReLU(0.2)(x)

    nf_mult, nf_mult_prev = 1, 1
    for n in range(n_layers):
        nf_mult_prev, nf_mult = nf_mult, min(2**n, 8)
        x = Conv2D(filters=ndf * nf_mult, kernel_size=(4, 4),
                   strides=2, padding='same')(x)
        x = BatchNormalization()(x)
        x = LeakyReLU(0.2)(x)

    nf_mult_prev, nf_mult = nf_mult, min(2**n_layers, 8)
    x = Conv2D(filters=ndf * nf_mult, kernel_size=(4, 4),
               strides=1, padding='same')(x)
    x = BatchNormalization()(x)
    x = LeakyReLU(0.2)(x)

    x = Conv2D(filters=1, kernel_size=(4, 4), strides=1, padding='same')(x)
    if use_sigmoid:
        x = Activation('sigmoid')(x)

    x = Flatten()(x)
    x = Dense(1024, activation='tanh')(x)
    x = Dense(1, activation='sigmoid')(x)

    model = Model(inputs=inputs, outputs=x, name='Discriminator')
    return model


def generator_containing_discriminator(generator, discriminator):
    image_shape = DeblurConfig.CELL_SHAPE
    inputs = Input(shape=image_shape)
    generated_image = generator(inputs)
    outputs = discriminator(generated_image)
    model = Model(inputs=inputs, outputs=outputs)
    return model


def generator_containing_discriminator_multiple_outputs(
        generator, discriminator):
    image_shape = DeblurConfig.CELL_SHAPE
    inputs = Input(shape=image_shape)
    generated_image = generator(inputs)
    outputs = discriminator(generated_image)
    model = Model(inputs=inputs, outputs=[generated_image, outputs])
    return model




from __future__ import division
from __future__ import print_function

import keras

from keras.models import Sequential
from keras.preprocessing.image import ImageDataGenerator
from keras import regularizers, optimizers
import numpy as np
import csv
import json
import os
from math import ceil
from collections import OrderedDict
import pandas as pd

from ....lib.model_lib import multi_label_classification_model_2
from ....lib.config_lib import MultiLabelClassificationConfig
from ....lib.generators import PredictGenerator
from ....lib.general_lib import get_n_largest_index_and_probs, get_data_path_type
from ....lib.saving_lib import save_models_with_class_indices, load_models_with_class_indices
from ....lib.image_lib import reshape_data

from ...application import Application


class Multi_Label_Image_Classification(Application):
    def __init__(self, model_path=None):
        super().__init__()

        if not isinstance(model_path, type(None)):
            self.load(model_path)

    def train(self,
            annotation_path=None,
            image_directory_path="",
            validation_data_proportion = 0.1,
            batch_size = 32,
            epochs = 1):        
        class_list, data_num = self.get_dataset_info(annotation_path = annotation_path)

        train_generator, valid_generator = self.make_train_valid_generator(annotation_path=annotation_path,
                                                                        image_directory_path=image_directory_path,
                                                                        batch_size=batch_size,
                                                                        class_list=class_list,
                                                                        validation_data_proportion=validation_data_proportion,
                                                                        data_num = data_num)

        self.model = self.create_and_compile_model(class_list = class_list)

        steps_train = ceil(train_generator.n / train_generator.batch_size)
        steps_valid = ceil(valid_generator.n / valid_generator.batch_size)

        self.model.fit_generator(generator = train_generator,
                            steps_per_epoch = steps_train,
                            validation_data = valid_generator,
                            validation_steps = steps_valid,
                            epochs = epochs)

    def save(self, model_path):
        save_models_with_class_indices(self.model, model_path)
        

    def load(self, model_path):
        self.model = load_models_with_class_indices(model_path)


    def predict(self, 
            data_path,
            verbose = 0,
            batch_size = 32,
            top_n = 3):
        
        data_path_type = get_data_path_type(data_path)
        if data_path_type == "directory":
            pred_datagen = PredictGenerator()
            pred_generator = pred_datagen.flow_one_directory(
                directory_path = data_path,
                shuffle = False,
                image_shape=MultiLabelClassificationConfig.INPUT_SHAPE[0:2],
                batch_size=batch_size
            )

            steps_pred = ceil(len(pred_generator.image_filenames) / pred_generator.batch_size)
            probs = self.model.predict_generator(pred_generator,
                                            steps = steps_pred,
                                            verbose = 1)

            predicted_classes_top_n, probs_top_n = get_n_largest_index_and_probs(probs, top_n)
            predicted_order_filenames = pred_generator.predict_order_filenames

            pair_dict = OrderedDict()
            for i in range(len(predicted_classes_top_n)):
                result_dict = OrderedDict()
                for j in range(0, top_n):
                    result_dict[self.model.class_indices[str(predicted_classes_top_n[i]
                                                    [j])]] = probs_top_n[i][j]
                pair_dict[predicted_order_filenames[i]] = result_dict
            return json.dumps(pair_dict, ensure_ascii=False)
        elif data_path_type == 'image':
            _, only_file_name = os.path.split(data_path)
            img_to_predict = cv2.imread(data_path)
            img_array = np.expand_dims(img_to_predict, axis=0)
            reshaped_array = reshape_data(data_array = img_array,
                                        input_shape = self.model.input_shape)
            probs = self.model.predict(reshaped_array, batch_size=batch_size, verbose=verbose, steps=steps)
            predicted_classes_top_n, probs_top_n = get_n_largest_index_and_probs(probs, top_n)

            result_dict = OrderedDict()
            for j in range(0, top_n):
                result_dict[self.model.class_indices[str(predicted_classes_top_n[0][j])]] = probs_top_n[0][j]
            pair_dict[only_file_name] = result_dict

            return json.dumps(pair_dict, ensure_ascii=False)


    def make_train_valid_generator(self,
                                annotation_path,
                                image_directory_path, 
                                batch_size,
                                class_list,
                                validation_data_proportion,
                                data_num):
        dataframe = pd.read_csv(annotation_path)

        class_list, dataframe = self.get_class_list_and_data_frame_from_csv(annotation_path)
        datagen = ImageDataGenerator(rescale = 1./255.)
        valid_datagen = ImageDataGenerator(rescale= 1./255.)

        train_data_num = (int) (data_num*(1-validation_data_proportion))

        train_generator = datagen.flow_from_dataframe(
            dataframe = dataframe[:train_data_num],
            directory = image_directory_path,
            x_col = "Filenames",
            y_col = class_list,
            class_mode = "other",
            target_size =MultiLabelClassificationConfig.INPUT_SHAPE[0:2],
            batch_size = batch_size
        )

        valid_generator = valid_datagen.flow_from_dataframe(
            dataframe = dataframe[train_data_num:],
            directory = image_directory_path,
            x_col = "Filenames",
            y_col = class_list,
            class_mode = "other",
            target_size =MultiLabelClassificationConfig.INPUT_SHAPE[0:2],
            batch_size = batch_size
        )

        return train_generator, valid_generator

    def create_and_compile_model(self, class_list):
        model = multi_label_classification_model_2(len(class_list), gpu_num = self.gpu_num)
        model.compile(optimizers.rmsprop(lr=0.001, decay=1e-6),loss="binary_crossentropy",metrics=["accuracy"])
        model.class_indices = {}
        for i in range(len(class_list)):
            model.class_indices[str(i)] = class_list[i]

        return model

        
    def get_dataset_info(self, annotation_path):
        class_list = []
        num_data = 0
        f = open(annotation_path, 'r', encoding='utf-8')

        csv_reader = csv.reader(f)
        class_row = next(csv_reader)

        for class_name in class_row:
            if class_name != "Filenames":
                class_list.append(class_name)
        
        for line in csv_reader:
            num_data += 1
            
        return class_list, num_data

            
    def get_class_list_and_data_frame_from_csv(self, csv_file_path):
        class_list = []
        dataframe = []
        with open(csv_file_path, 'r', encoding='utf-8') as f:
            csv_reader = csv.reader(f)
            class_row = next(csv_reader)

            for class_name in class_row:
                class_list.append(class_name)

            for line in csv_reader:
                probs_list = line[1:]
                dataframe.append(probs_list)

        dataframe = np.array(dataframe)
        return class_list, dataframe
            
            

